/****************************************************************************
ipwrij.cpp

Tue 6/12/2007 5:11 pm     creation.

Includes RIJNDAEL-ALG-FST.C et al (below), where they say

 * Optimised ANSI C code for the Rijndael cipher (now AES)
 *
 * @author Vincent Rijmen <vincent.rijmen@esat.kuleuven.ac.be>
 * @author Antoon Bosselaers <antoon.bosselaers@esat.kuleuven.ac.be>
 * @author Paulo Barreto <paulo.barreto@terra.com.br>
 *
 * This code is hereby placed in the public domain.

THE STORY OF IPW ENCRYPTION, Thu 6/14/2007 3:24 pm

I will tell it here, writing in the sand for whoever will read. ... When I
begin the IPW experiment, I sought-out free encryption code on the web, and
came upon what seemed like appropriate stuff which I incorporated with great
difficulty into IPW.

But I was betrayed by Wikipedia! They told me the stuff was "public domain".
In fact, the author had included *no* copyright, only a name and email address
which, sadly, is nothing; copyright sticks with the guy who wrote it, whether
he says or no, *unless* he says something -- which is why *this* IPW code is
released specially under the BSD license (see control_ipw.mm.).

Of course it wasn't wikipedia's fault, but mine, and I was punished by being
forced to figure-out this nonsense *a second time*, for the admirable source
code noted above and included below. Remnants of this process in the source
and particularly IPW.H support running *both* kinds of encryption for a while
(RIJ_ENC0) ... but this is finished. "ENC0" the original dubious-copyright
encryption is no more, and the RIJNDAEL is all there is.

----------------------------------------
THIS FILE: THE PLOT, Tue 8/14/2007 4:25 pm

I see I've left-out earlier comments of what this file is up-to. The general
idea is I tried to implement this encryption stuff with a class or something,
and it got just too hairy -- the included C code you know -- so what I do is
use this file, which has numerous statics/globals, and presumably a few entry
points -- the prototypes for which are defined in ipw.h (that circumstance
seems to be related to the brief dual-running history referred to above).

Anyway, with both ENC0 and the superior RIJNDAEL I adopted a strategy of
taking the C demo program typically supplied with these things, and
factoring-out the i/o -- i.e., typically the file character-in/character-out
-- into a function like

  static void rij_io_decrypt (const char *mykey, ICRYP_FUNC in_c, OCRYP_FUNC
  out_c) {

where I/OCRYP_FUNC is defined in ipw.h like

  typedef int (*OCRYP_FUNC)(unsigned char);
  typedef int (*ICRYP_FUNC)(void);

and then call rij_io_decrypt like

  rij_io_decrypt(akey,rij_from_file_deasciify,to_list);

where rij_from... and to_list are in here, and perform the high-toned C++
style activities.

This let me use encryption code which I frankly don't understand in a fairly
contained way.

j.g. owen * * * * * * * * * * * * * * * * * * *
web:   http://owenlabs.home.att.net/
email: owen_bda4@yahoo.com
* * * * * * * * * * * * * * * * * * * * * * * *

****************************************************************************/
#include "ipw.h"

#if RIJNDAEL  //entire module.

/****************************************************************************
****************************************************************************/


#define ipwJGO (1)

#include "rijndael-alg-fst.c"
#include "rijndael-api-fst.c" //apparently it calls the previous.

/****************************************************************************
****************************************************************************/


//#define THEKEY "we'll use the same key"
extern char THEKEY[];

static int rij_crypt_ascii_count;

/****************************************************************************
****************************************************************************/
static void rij_io_decrypt (const char *mykey, ICRYP_FUNC in_c, OCRYP_FUNC
out_c) {
  keyInstance k;
  cipherInstance cipher;
  int i, c, count=0;
  BYTE in_array[16], out_array[16];

  rij_crypt_ascii_count = 0;  //not used here I suppose.

  if (makeKey(&k,DIR_DECRYPT,128,mykey)<0) {
    printf("rij_io_encrypt: bad makeKey\n");
    exit(17);
  }

  /****************************************************************
  in enc0.c

    memcpy (expkey, key, Nk * 4); //4 * 4 ==> 16.

  and 16 * 8 ==> 128 bits. Or something.
  ****************************************************************/

//  cipher.blockLen = 128;  //apparently none of that.

  ///             ??     ??
  cipherInit(&cipher,MODE_ECB,NULL);
  //the other choices are MODE_CBC MODE_CFB1

  while ( (c=in_c()) != EOF) {
    in_array[count++] = c;
    if (count<16)
    continue;

    count = 0;
//.... so down here somewhere it does something like this.
    blockDecrypt(&cipher,&k,in_array,128,out_array);
    for (i = 0; i<16; ++i) {
      out_c(out_array[i]);
    }
  }

#if SHOULDNT_HAPPEN
  if (count) {
    for (i=count; i<16; ++i)  //zero-out the rest.
    in_array[i]=0;

    for (i=0; i<count; ++i)
    out_c(out_array[i]
  }
#endif

}


static void rij_io_encrypt (const char *mykey, ICRYP_FUNC in_c, OCRYP_FUNC out_c) {
  keyInstance k;
  cipherInstance cipher;
  int i, c, count=0;
  BYTE in_array[16], out_array[16];

  rij_crypt_ascii_count = 0;

  if (makeKey(&k,DIR_ENCRYPT,128,mykey)<0) {
    printf("rij_io_encrypt: bad makeKey\n");
    exit(17);
  }

#if JGO
//  setmem(&cipher,sizeof(cipherInstance),0);
#endif

//  cipher.blockLen = 128;

  ///             ??     ??
  cipherInit(&cipher,MODE_ECB,NULL);
  //the other choices are MODE_CBC MODE_CFB1

  while ( (c=in_c()) != EOF) {
    in_array[count++] = c;
    if (count<16)
    continue;

    count = 0;
//.... so down here somewhere it does something like this.
    blockEncrypt(&cipher,&k,in_array,128,out_array);
    for (i = 0; i<16; ++i) {
      out_c(out_array[i]);
    }
  }

  if (count) {
    for (i=count; i<16; ++i)  //zero-out the rest.
    in_array[i]=0;

    blockEncrypt(&cipher,&k,in_array,128,out_array);

    for (i=0; i<16; ++i)
    out_c(out_array[i]);
  }

}


/****************************************************************************
****************************************************************************/

static FILE *i_crypt_file;
static FILE *o_crypt_file;

static int rij_from_file(void) {
  return
  fgetc(i_crypt_file);
}

static int rij_to_file(unsigned char c) {
  return
  fputc(c,o_crypt_file);
}

/****************************************************************************
discards zeros; presumably artifacts of encryption process. This is quite
crude, since it could discard legitmate zeros in some wacko coding scheme. But
I'm pretty-much assuming ASCII text, at least ASCII-enough to exclude chr$(0).

The way to fix that would be to mysteriously encode zeros on the way in with
an escape character etc. I might eventually bother doing this in connection
with preserving line-endings properly somehow maybe....
****************************************************************************/

static int rij_to_file_nz(unsigned char c) {
  if (c)
  return
  fputc(c,o_crypt_file);
  return 0;
}

/****************************************************************************
Note that Goptions.dont_ascii_encrypt refers to saving of encrypted material;
if Goptions.dont_ascii_encrypt, then it's saved as binary data; else it's
saved in a crudely-encoded ascii. To avoid encryption entirely, use the
"crypt" argument to list_encrypt_file, file_decrypt_list (i.e.
programatically).
****************************************************************************/

//asciify c on the way out; i.e. this is called by encrypt code to send a
//character, and we asciify it before putting it in file.
static int rij_to_file_asciify(unsigned char c) {

#if 0
  if (Goptions.dont_ascii_encrypt)
  return fputc(c,o_crypt_file);
#endif

  if (
  (fputc( (c / 16)+'a', o_crypt_file)==EOF) ||
  (fputc( (c & 15)+'a', o_crypt_file)==EOF)
  )
  return EOF;

  if ((rij_crypt_ascii_count+=2) > 75) {
    rij_crypt_ascii_count = 0;
    if (fputc('\n',o_crypt_file)==EOF)
    return EOF;
  }

  return 0;
}

//get 2 characters from file, and deasscify them.
static int rij_from_file_deasciify(void) {
  int c1, c2;

#if 0
  if (Goptions.dont_ascii_encrypt)
  return fgetc(i_crypt_file);
#endif

  do {
    if ( (c1=fgetc(i_crypt_file))==EOF)
    return EOF;
  } while (c1<'a');       //throw away non 'a's like line returns.

  do {
    if ( (c2=fgetc(i_crypt_file))==EOF)
    return EOF;
  } while (c2<'a');       //throw 'em away!
  //see note above re to_file_nz.

  return
  (c1-'a') * 16
  + (c2-'a')
  ;


}

/****************************************************************************
the binary versions.
****************************************************************************/

static int rij_test_e(char *iname, char *crypt_name) {
  mustopen(&i_crypt_file,iname,"r");
  mustopen(&o_crypt_file,crypt_name,"wb");
  rij_io_encrypt(THEKEY,rij_from_file,rij_to_file);
  fclose(i_crypt_file);
  fclose(o_crypt_file);

  return 0;

}

static int rij_test_d(char *crypt_name, char *o_name) {
  mustopen(&i_crypt_file,crypt_name,"rb");
  mustopen(&o_crypt_file,o_name,"w");
  rij_io_decrypt(THEKEY,rij_from_file,rij_to_file_nz);
  fclose(i_crypt_file);
  fclose(o_crypt_file);

  return 0;

}

/****************************************************************************
****************************************************************************/

int rij_test_e_d(char *iname, char *crypt_name, char *o_name) {
  rij_test_e(iname,crypt_name);
  rij_test_d(crypt_name,o_name);

  return 0;
}

/****************************************************************************
the ascii versions.
****************************************************************************/

static int rij_test_e_a(char *iname, char *crypt_name) {
  mustopen(&i_crypt_file,iname,"r");
  mustopen(&o_crypt_file,crypt_name,"w");
  rij_io_encrypt(THEKEY,rij_from_file,rij_to_file_asciify);
  fclose(i_crypt_file);
  fclose(o_crypt_file);

  return 0;

}

static int rij_test_d_a(char *crypt_name, char *o_name) {
  mustopen(&i_crypt_file,crypt_name,"r");
  mustopen(&o_crypt_file,o_name,"w");
  rij_io_decrypt(THEKEY,rij_from_file_deasciify,rij_to_file_nz);
  fclose(i_crypt_file);
  fclose(o_crypt_file);

  return 0;

}

int rij_test_e_d_a(char *iname, char *crypt_name, char *o_name) {
  rij_test_e_a(iname,crypt_name);
  rij_test_d_a(crypt_name,o_name);

  return 0;
}

/****************************************************************************
****************************************************************************/
static JPWBUNCH::iterator jpwi;

static string r_astring, r_ostring;
static string::iterator r_astring_i;

static JSTRINGLIST r_astringlist;
static JSTRINGLIST::iterator r_alist_i;

static int jcrypt_count;

static void clearjunk(void) {

//  Gthelist.clear();      //nope this is the global list.
  jpwi = Gthelist.end();

  r_astring.clear();
  r_ostring.clear();

  r_astring_i = r_astring.end();

  r_astringlist.clear();
  r_alist_i = r_astringlist.end();

  from_list_count =
  jcrypt_count =
  rij_crypt_ascii_count = 0;

}

/****************************************************************************
****************************************************************************/

//return character-by-character, list in jpwlist.
static int from_list(void) {

  if (r_astring_i == r_astring.end()) {  //get another thingey.

    if (jpwi == Gthelist.end())
    return EOF;                     //we are finished.

    //else make r_astring into contents of *jpwi.
    r_astringlist = jpwi->jpw_ascii();
    ++jpwi;                         //next record.

    ++from_list_count;              //count something.

    //now, concatenate the string list onto the single r_astring. there's
    //probably a right way to do this -- but maybe not.
    r_astring.clear();
    for (r_alist_i=r_astringlist.begin();
    r_alist_i!=r_astringlist.end(); ++r_alist_i) {
      r_astring += *r_alist_i + '\n';
    }

    r_astring_i = r_astring.begin();
  }

//and now return the next character in the string.
  return *r_astring_i++;

}

/****************************************************************************
****************************************************************************/

//store, character-by-character, into list jpwlist.
static int to_list(unsigned char c) {
  if (c=='\n') {      //do something.
    r_astringlist.push_back(r_astring);
    if (r_astring.substr(0,tLENGTH)==tCOMM) { //the terminator.
      jpw ajpw;
      ajpw.ascii_jpw(r_astringlist);
      Gthelist.push_back(ajpw);    //add a record.
      r_astringlist.clear();        //clear the accumulating string list.
    }
    r_astring.clear();              //in any case, clear the string.

  }
  else
  r_astring += c;   //else concatenate a character onto the string.


  return 0;
}

/****************************************************************************
NOTE: we do the non-crypt versions.

****************************************************************************/
static void io_dont(ICRYP_FUNC in_c, OCRYP_FUNC outc) {
  int c;
  while((c=in_c())!=EOF)
  outc(c);

}

/****************************************************************************
****************************************************************************/
static int from_string(void) {
  if (r_astring_i == r_astring.end())
  return EOF;

  return
  *r_astring_i++;
}

static int to_string(unsigned char c) {
  r_ostring += (c / 16)+'a';
  r_ostring += (c & 15)+'a';
  return 0;
}



/****************************************************************************
****************************************************************************/

string rij_encrypt_string(const string &s) {
  char *akey;

  akey = strdup(s.c_str());
  if (Goptions.debug)
  printf("rij_encrypt_string(%s)\n",akey);

  clearjunk();

  r_astring = s;
  r_astring_i = r_astring.begin();

  rij_io_encrypt(akey,from_string,to_string);
//we ecrypt with the string as a key?

  free(akey);

  if (Goptions.debug)
  printf("rij_encrypt_string:%s\n",r_ostring.c_str());

  return r_ostring;
}

bool rij_list_encrypt_file(const char *akey, const char *crypt_name, bool
crypt) {
  clearjunk();

  jpwi = Gthelist.begin();

#if j_WIN
  if (Goptions.dont_ascii_encrypt)
  o_crypt_file = fopen(crypt_name,"wb");  //must be BINARY without ascii!
  else
#endif
  o_crypt_file = fopen(crypt_name,"w");

  if (!o_crypt_file) return
  false;

  if (crypt)
  rij_io_encrypt(akey,from_list,rij_to_file_asciify);
  else
  io_dont(from_list,rij_to_file);

  fclose(o_crypt_file);

  return true;


}

bool rij_file_decrypt_list(const char *akey, const char *crypt_name, bool crypt) {
  clearjunk();

#if j_WIN
  if (Goptions.dont_ascii_encrypt)
  i_crypt_file = fopen(crypt_name,"rb");
  else
#endif
  i_crypt_file = fopen(crypt_name,"r");

  if (!i_crypt_file)
  return false;

  Gthelist.clear();

  if (crypt)
  rij_io_decrypt(akey,rij_from_file_deasciify,to_list);
  else
  io_dont(rij_from_file,to_list);

  fclose(i_crypt_file);

  return true;
}

/****************************************************************************
****************************************************************************/
#endif  //if RIJNDAEL entire module.
