/****************************************************************************
jini.cpp
Fri 1/26/2007 2:39 pm   creation.
Tue 5/01/2007 11:09 am  added read() q.v.
Wednesday, May 9, 2007 1:25 pm  Hopefully-harmless changes to Visual Studio
                        2005 version.
Wednesday, May 16, 2007 12:41 pm  Nasty bugs.
Wed 5/30/2007 9:47 am   made constructor/destructor agree on argument types.
Compilers didn't seem to mind difference.

Fri 6/15/2007 5:08 pm   Added readlist et al.

This would be the infamous "James gregor owen" "INI" file code; the idea is to
provide a Delphi-like ini gadget one can read windows-style ini file options
from.

****************************************************************************/
#include "jini.h"

namespace jIni_space {

/****************************************************************************
****************************************************************************/

jIni::jIni(const char *name) {
  filename = name;
  read_file();
}

jIni::jIni(void) {
}

jIni::~jIni(void) {
}

/****************************************************************************
****************************************************************************/
void jIni::read(const char *name) {
  changename(name);
  read_file();
}

/****************************************************************************
Wed 1/31/2007 3:08 pm. very handy for testing.....
****************************************************************************/
void jIni::changename(const char *name) {
  filename = name;
}

/****************************************************************************
****************************************************************************/
std::string jIni::readstr(const char *sec, const char *key, const char *def) {
  std::string val = find_value(sec,key);

  if (!val.length())
  val = def;

  return val;
}


char *jIni::readc_str(const char *sec, const char *key, const char *def) {
  std::string val = find_value(sec,key);

  if (!val.length())
  return strdup(def);

  return strdup(val.c_str());

}

int jIni::readint(const char *sec, const char *key, int def) {

  return
  StringToInt(readstr(sec,key,IntToString(def).c_str() ) );

}



bool jIni::readbool(const char *sec, const char *key, bool def) {
  std::string val = find_value(sec,key);

  if (!val.length())
  return def;

  return (val!="0");

}


/****************************************************************************
****************************************************************************/

void jIni::writestr(const char *sec, const char *key, std::string val) {
  JSTRINGLIST::iterator i = find_value_it(sec,key);

  if (i==lines.end()) {
    i = find_section(sec);
    if (i==lines.end()) {
      lines.insert(lines.end(),std::string("[") + sec + "]");
      i = lines.insert(lines.end(),"");
    }
    else
    i = lines.insert(i,"");

  }

  *i = std::string(key) + "=" + val;

}

void jIni::writeint(const char *sec, const char *key, int val) {
  writestr(sec,key,IntToString(val));
}

void jIni::writebool(const char *sec, const char *key, bool val) {
  std::string s("0");
  if (val)
  s = "1";

  writestr(sec,key,s);

}

/****************************************************************************
****************************************************************************/


//writes it with any changes, NOT automatically.
void jIni::writeini(void) {
  write_file();
}

/****************************************************************************
****************************************************************************/
void jIni::read_file(void) {
  std::string s;
  std::ifstream ifile(filename.c_str() /* ! */ );

//just fill-up the lines with the file. Seems like there ought to be a
//simpler way?
  if (ifile /* .is_open() */){
    lines.clear();
    while (ifile /*!ifile.eof()*/ ) {
      getline(ifile,s);
      if (ifile) {    //can't tell if finished until getline.
        decontrolM(s);
        lines.push_back(s);
      }
    }
  }
}

void jIni::write_file(void) {
  JSTRINGLIST::iterator i;
  std::ofstream ofile(filename.c_str());
  if (ofile) {
    i = lines.begin();
    while(ofile && (i!=lines.end())) {
      ofile << *i++ <<  std::endl;    // no worko: << "/n"
//      ++i;
    }
  }
}

/****************************************************************************
****************************************************************************/
JSTRINGLIST::iterator jIni::find_section(const char *section) {
  JSTRINGLIST::iterator i;
  std::string keyname = std::string("[") + section + "]";
  size_t keylength=keyname.length();

  for (i=lines.begin(); i!=lines.end(); ++i) {
    if (jstringeq((i->substr(0,keylength)),keyname))
    return ++i;   //the first value.
  }

  return lines.end();  //means nada?

}


std::string jIni::find_value(const char *section, const char *key) {
  std::string keyname = std::string(key) + "=";
  size_t keylength=keyname.length();
  JSTRINGLIST::iterator i=find_section(section);

  for (; i!=lines.end(); ++i) {
//check for next section.
    if (i->c_str()[0]=='[')
    return "";

    if (jstringeq((i->substr(0,keylength)),keyname))
    return i->substr(keylength); //stuff to right of =.
  }

  return "";

}

JSTRINGLIST::iterator jIni::find_value_it(const char *section, const char
*key) {
  std::string keyname = std::string(key) + "=";
  size_t keylength=keyname.length();
  JSTRINGLIST::iterator i=find_section(section);

  for (; i!=lines.end(); ++i) {
    if (jstringeq((i->substr(0,keylength)),keyname))
    return i;
  }

  return lines.end();

}

/****************************************************************************
so he gets the section into the map, and then fiddles with it. However he
has to free it here.
****************************************************************************/
INIMAP *jIni::readsec(const char *sec) {
  INIMAP *m = new INIMAP;
  int c;
  size_t x;
  std::string key;
  mapval mv;

  JSTRINGLIST::iterator i = find_section(sec);

  for (; i!=lines.end(); ++i) {
//check for next section.

    if ( (c=i->c_str()[0]) =='[')
    return m;

    if (ispunct(c) || (c<=' ') )
    continue;   //skip comments and any other nonsense.

    if ((x = i->find('=',0))== std::string::npos)
    continue;

    key = strdup(i->substr(0,x).c_str());
    mv.s = strdup(i->substr(x+1,1000).c_str());
    mv.v = atoi(mv.s);

//    m->insert(std::pair<std::string, mapval>(key,mv));
    m->insert(PAIRMAP(key,mv));

  }

  return m;

}

void jIni::writesec(const char *sec, INIMAP *m) {
  INIMAP::iterator i;
  for ((i=m->begin()); i!=m->end(); ++i) {

//just write the sections as usual!
    writestr(sec,i->first.c_str(),i->second.s);

  }

}

void jIni::freesec(INIMAP **map) {
  INIMAP *m = *map;
  INIMAP::iterator i;

  *map = NULL;

  if (!m)
  return;

  for (i=m->begin(); i!=m->end(); ++i) {
    freeandnil(&(i->second.s));
  }

  delete m;
  //I foolishly assume it knows about its own thingeys i.e. i->first.

}

/****************************************************************************
****************************************************************************/
JSTRINGLIST *jIni::readlist(const char *sec) {
  JSTRINGLIST *j = new JSTRINGLIST;
  JSTRINGLIST::iterator i = find_section(sec);
  size_t x;
  int c;
  std::string key;

  for (; i!=lines.end(); ++i) {
//check for next section.

    if ( (c=i->c_str()[0]) =='[')
    return j;

    if (ispunct(c) || (c<=' ') )
    continue;   //skip comments and any other nonsense.

    if ((x = i->find('=',0))== std::string::npos)
    continue;   //no equals sign, no symbol.

    if ( (i->substr(x+1,1))=="1") //gotta equal 1 to be used.
    j->push_back(i->substr(0,x));

  }

  return j;

}

void jIni::writelist(const char *sec, JSTRINGLIST *list) {
  JSTRINGLIST::iterator i;

  if (!list)
  return;

  for (i=list->begin(); i!=list->end(); ++i) {
    writestr(sec,i->c_str(),"1");
  }

}

//or you could just delete it.
void jIni::freelist(JSTRINGLIST **list) {
  if (*list)
  delete *list;

  *list = NULL;
}

/****************************************************************************
****************************************************************************/
} //namespace jIni_space.
/****************************************************************************
****************************************************************************/


